/* This file has a simple parser that acts as an interface */
/* to the generic steensgaard routines.                    */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "node.h"
#include "union_find.h"
#include "steensgaard.h"

/* Cheating, and using global vars */
static int ch;
int line=0;
int dump_html=0;
char *html_prefix=NULL,temp_string[BUFSIZ];
FILE *html_out=NULL;

#define VERSION "0.8.0"

/* These should probably be enums */
#define PREFIX_NONE 0
#define PREFIX_STAR 1
#define PREFIX_AMP  2

#define VAR_NONE     0
#define VAR_VARIABLE 1
#define VAR_FUNCTION 2
#define VAR_NUMBER   3


   /* skips whitespace in the parser */
   /* also handles // comments       */
   /* returns -1 if EOF              */
int skip_whitespace(FILE *input) {
      
   while(1) {
      
         /* watch for EOF */
      if (ch<0) return -1;
      
         /* Handle comments */
      if (ch=='/') {
         ch=fgetc(input);
	 if (ch=='/') {
	    while((ch!='\n') && (ch>=0)) ch=fgetc(input);
	    line++;
	 } 
	 else {
	    return 0;
	 }	 
      }
      
         /* Increment line count */
      if (ch=='\n') line++;

         /* No longer whitespace, return */
      if (ch>' ') return 0;

      ch=fgetc(input);      
   }

      /* We should never get here */
   return -1;
}


    /* Parser part that looks for * or & in front of a var */
    /* Doesn't handle multiple * or & characters           */
int get_prefix(FILE *input) {
      
      if (ch=='*') {
	 ch=fgetc(input);
	 return PREFIX_STAR;
      }
      
      if (ch=='&') {
	 ch=fgetc(input);
	 return PREFIX_AMP;
      }
      
         /* EOF */
      if (ch<0) return -1;
      
      return PREFIX_NONE;
}

    /* Gets a variable.  If it doesn't exist, create it */
    /* Returns what type of var, or returns -1 if EOF   */
int get_var(FILE *input, struct node **var) {

   char temp_name[BUFSIZ];
   int offset=0;
   
      /* Handle EOF */
   if (ch<0) return -1;
   
      /* Handle if it's a number instead of a var */
   if (isdigit(ch)) {
      while(1) {
	 ch=fgetc(input);
	 if (ch<0) return -1;
	 if (!isdigit(ch)) break;
      }
      return VAR_NUMBER;
   }
   
   temp_name[offset]=ch;
   offset++;
   
   while(1) {
      
      ch=fgetc(input);
      
         /* handle EOF */
      if (ch<0) return -1;
      
         /* handle if variable name is too long */
      if (offset>BUFSIZ) {
	 printf("Error!  Variable name too long at %i\n",line);
	 return -3;
      }
      
         /* variable name can start with alpha or _  */
         /* rest of variable can be alpha, num, or _ */
      if (!isalpha(ch)&&!isdigit(ch)&&ch!='_') {
	 temp_name[offset]='\0';
	 
	    /* if it ends in ()  it is a function */
	 if (ch=='(') {
	    ch=fgetc(input);
	    if (ch==')') {
	       ch=fgetc(input);
	       return VAR_FUNCTION;
	    }
	    else {
	       printf("Unknown partial function line %i\n",line);
	    }
	 }
	 	 
	    /* turn string into a var.  allocate if not there yet */
	 *var=find_node(temp_name);
	 if (*var==NULL) *var=new_node(temp_name);
	 return VAR_VARIABLE;
      }
      else {
         temp_name[offset]=ch;
	 offset++;
      }
   }
}

   /* Look for an equals sign.  Also supports Pascal type := equals */
int check_equals(FILE *input) {

      /* handle := */
   if (ch==':') {
      ch=fgetc(input);
      if (ch!='=') {
	 printf("Error!  Expected equals line %i\n",line);
	 return -2;
      }
   }
   
      /* handle = */
   if (ch=='=') {
      ch=fgetc(input);
      return 0;
   }
   
   return -1;
}


   /* Look for semicolon at the end of a line */
int check_semicolon(FILE *input) {

   if (ch!=';') {
      printf("Error!  Semi-colon expected at line %i\n",line);
      return -1;
   }

   ch=fgetc(input);
   
   return 0;
}

   /* Print the help message */
int dump_help(char *exe,int print_help) {
   
   printf("\nSteensgaard tool version %s\n",VERSION);
   printf("\tby Vince Weaver <vince _at_ csl.cornell.edu>\n\n");
   if (print_help) {
      printf("\t%s -debug -show-dummy -h -html=prefix\n\n",exe);
      printf("\t-debug       = turn on debug messages\n");
      printf("\t-show-dummy  = show dummy variables\n");
      printf("\t-h           = print this help\n");
      printf("\t-html=PREFIX = create HTML result, with prefix given\n");
   }
   printf("\n");
   exit(0);   
}


   /* Print the assignment that just happened    */
   /* Also print to HTML if we have that enabled */
void dump_line(char *left_prefix,struct node *left_var,
	       char *equals,
	       char *right_prefix,struct node *right_var) {

   static int line_count=0;
   
      /* print to html file */
   if (dump_html) {
      fprintf(html_out,"<HR>\n");
      if (left_prefix!=NULL) fprintf(html_out,left_prefix); 
      dump_node_name(html_out,left_var);  
      fprintf(html_out,equals);
      if (right_prefix!=NULL) fprintf(html_out,right_prefix);
      if (right_var==NULL) {
         fprintf(html_out,"bottom");
      } 
      else {
         dump_node_name(html_out,right_var); 
      }
      
      dump_nodes_html(html_prefix,line_count);
      fprintf(html_out,"<BR>\n");	    
      fprintf(html_out,"<CENTER><IMG SRC=\"%s_%i.png\"></CENTER><BR>\n",
	      html_prefix,line_count);
      line_count++;
      
   }
   
      /* print to stdout */
   if (left_prefix!=NULL) printf(left_prefix); 
   dump_node_name(stdout,left_var);  
   printf(equals);
   if (right_prefix!=NULL) printf(right_prefix);
   if (right_var==NULL) {
      printf("bottom");
   } 
   else {
      dump_node_name(stdout,right_var); 
   }
   printf("\n");	    
   
}

int main(int argc, char **argv) {

   int result,i;
   struct node *left_var,*right_var;
   int left_prefix,right_prefix,left_type,right_type;
   
   /* parse command line arguments */
   
   for(i=1;i<argc;i++) {
         /* Turn on Debugging */
      if (!strcmp(argv[i],"-debug")) {
	 debug_messages=1;
      }
         /* Enable printing of dummy variables */
      if (!strcmp(argv[i],"-show-dummy")) {
	 print_dummy=1;
      }
         /* Print help messages */
      if (!strcmp(argv[i],"-h")) {
	 dump_help(argv[0],1);
      }
         /* Print version */
      if (!strcmp(argv[i],"-v")) {
	 dump_help(argv[0],0);
      }
         /* Output to html file */
      if (!strncmp(argv[i],"-html=",6)) {
	 dump_html=1;
	 html_prefix=strdup(argv[i]+6);
	 sprintf(temp_string,"%s.html",html_prefix);
	 html_out=fopen(temp_string,"w");
	 if (html_out==NULL) {
	    printf("Error!  Can't open file %s\n",temp_string);
	    exit(1);
	 }
	 fprintf(html_out,"<HTML>\n");
	 fprintf(html_out,"<HEAD><TITLE>Steensgaard for %s</TITLE></HEAD>\n",
		 html_prefix);
	 fprintf(html_out,"<BODY>\n");
	 fprintf(html_out,"<CENTER><H3>Steensgaard for %s</H3></CENTER>\n",
		 html_prefix);
	 fprintf(html_out,"<HR>\n");
	 fprintf(html_out,"Input that generated this file: ");
	 fprintf(html_out,"<A href=\"%s.test\">%s.test</A>\n",
		 html_prefix,html_prefix);
	 
      }

   }
   
   
      /* Loop Forever */
   while(1) {

         /* clear the variables */
      left_var=NULL;  right_var=NULL;
      left_prefix=PREFIX_NONE;
      right_prefix=PREFIX_NONE;
      left_type=VAR_NONE;
      right_type=VAR_NONE;
      
         /* skip leading whitespace */
      result=skip_whitespace(stdin);
      if (result<0) break;

         /* get left prefix, if there is one */
      left_prefix=get_prefix(stdin);
      if (left_prefix<0) break;

         /* skip whitespace */
      result=skip_whitespace(stdin);
      if (result<0) break;
  
         /* if a variable or function, get variable */
      if (isalpha(ch)||ch=='_') {
	 
	    /* get variable */
         left_type=get_var(stdin,&left_var);      
         if (left_type<0) break;

	    /* If it's a function, call dump_nodes() and done with line */
	 if (left_type==VAR_FUNCTION) {
	    dump_nodes(stdout);
	    goto early_out;
	 }
	    /* skip whitespace */
         result=skip_whitespace(stdin);
         if (result<0) break;
      
	    /* look for equals sign */
         result=check_equals(stdin);
         if (result<0) break;
             
	    /* skip whitespace */
         result=skip_whitespace(stdin);
         if (result<0) break;
      }
      
         /* get the right-hand prefix, if any */
      right_prefix=get_prefix(stdin);
      if (right_prefix<0) break;
      
         /* skip whitespace */
      result=skip_whitespace(stdin);
      if (result<0) break;
      
         /* get the right hand variable, function, or number */
      right_type=get_var(stdin,&right_var);
      if (right_type<0) break;
         
         /* skip whitespace */
      result=skip_whitespace(stdin);
      if (result<0) break;

early_out:      
      
         /* look for end-of-line semicolon */
      result=check_semicolon(stdin);
      if (result<0) break;

         /* If we had a function on right, dump the nodes */
      if (right_type==VAR_FUNCTION) {
	 dump_nodes(stdout);
      }
      
         /* if right-hand was a number, then we union with bottom */
         /* which is a nop */
      if ((left_type==VAR_VARIABLE) && (right_type==VAR_NUMBER)) {
	 dump_line(NULL,left_var,"=",NULL,NULL);
      }
      
         /* We have an operation to do */
      if ((right_type==VAR_VARIABLE) && (left_type==VAR_VARIABLE)) {
	 
	    /* a=b */
	 if ((left_prefix==PREFIX_NONE) && (right_prefix==PREFIX_NONE)) {
	    equals(left_var,right_var);
	    dump_line(NULL,left_var,"=",NULL,right_var);
	 }
	    /* a=*b */
	 else if ((left_prefix==PREFIX_NONE) && (right_prefix==PREFIX_STAR)) {
	    equals_star(left_var,right_var);
	    dump_line(NULL,left_var,"=","*",right_var);
	 }
	    /* a=&b */
	 else if ((left_prefix==PREFIX_NONE) && (right_prefix==PREFIX_AMP)) {
	    equals_amp(left_var,right_var);
	    dump_line(NULL,left_var,"=","&",right_var);

	 }
	    /* *a=b */
	 else if ((left_prefix==PREFIX_STAR) && (right_prefix==PREFIX_NONE)) {
	    star_equals(left_var,right_var);
	    dump_line("*",left_var,"=",NULL,right_var);	    
	 }
	 else {
	    printf("Error!  Unhandled pointer operation at line %i\n",line);
	    break;
	 }
      }
   }

      /* close out the HTML file */
   if (dump_html) {
      fprintf(html_out,"\n<HR></BODY></HTML>\n");
      fclose(html_out);
   }
   
   return 0;
}

     
   
